/*
 *
 * Difficulty (Updated)
 *
 */

const events = require('events');
const utils = require('./utils');

////////////////////////////////////////////////////////////////////////////////

// Main RingBuffer Function
const RingBuffer = function(maxSize) {

  let data = [];
  let cursor = 0;
  let isFull = false;

  // Append to Ring Buffer
  this.append = function(x) {
    if (isFull) {
      data[cursor] = x;
      cursor = (cursor + 1) % maxSize;
    } else {
      data.push(x);
      cursor += 1;
      if (data.length === maxSize) {
        cursor = 0;
        isFull = true;
      }
    }
  };

  // Average Ring Buffer
  this.avg = function() {
    const sum = data.reduce((a, b) => a + b);
    return sum / (isFull ? maxSize : cursor);
  };

  // Calculate Size of Ring Buffer
  this.size = function() {
    return isFull ? maxSize : cursor;
  };

  // Clear Ring Buffer
  this.clear = function() {
    data = [];
    cursor = 0;
    isFull = false;
  };
};

// Main Difficulty Function
const Difficulty = function(port, difficulty) {

  const _this = this;
  _this.options = difficulty;

  let lastTs, timeBuffer;

  // Update Difficulty on Share Submission
  this.updateDifficulty = function(client) {
    const ts = (Date.now() / 1000) | 0;

    if (!lastTs) {
      lastTs = ts;
      timeBuffer = new RingBuffer(_this.options.targetSample);
      return;
    }

    const sinceLast = ts - lastTs;
    timeBuffer.append(sinceLast);
    lastTs = ts;
    const avg = timeBuffer.avg();
    const coef = _this.options.targetTime / avg;

    // Not enough sample to know, keep sampling
    if(timeBuffer.size() < _this.options.targetSample) {
        return;
    }

    // Clear the buffer
    timeBuffer.clear();

    // Bail-out if we're good
    if((coef < (1 + _this.options.variance)) && (coef > (1 - _this.options.variance))) {
        return;
    }

    // We're out of "the zone", let's try to fix that
    // That "should"(tm) be the right new difficulty
    const newDiff = utils.toFixed(client.difficulty * coef, 8);

    // Ensure we don't go higher than legal maximum and minimum
    // Also make sure we don't swing to hard after the initial correction
    if(newDiff > _this.options.maximum) {
        _this.emit('newDifficulty', client, _this.options.maximum);
    } else if(newDiff < _this.options.minimum) {
        _this.emit('newDifficulty', client, _this.options.minimum);
    } else if(newDiff > (client.difficulty * (1 + _this.options.variance)) && client.difficulty != _this.options.initial) {
        _this.emit('newDifficulty', client, utils.toFixed(client.difficulty * (1 + _this.options.variance), 8));
    } else if(newDiff < (client.difficulty * (1 - _this.options.variance)) && client.difficulty != _this.options.initial) {
        _this.emit('newDifficulty', client, utils.toFixed(client.difficulty * (1 - _this.options.variance), 8));
    } else {
        _this.emit('newDifficulty', client, newDiff);
    }
  };

  // Manage Individual Clients
  this.manageClient = function(client) {
    const stratumPort = client.socket.localPort;
    if (stratumPort != port) {
      console.error('Handling a client which is not of this vardiff?');
    }
    client.on('submit', () => _this.updateDifficulty(client));
  };
};

module.exports = Difficulty;
Difficulty.prototype.__proto__ = events.EventEmitter.prototype;
